/*
** UC: 21111 - Sistemas Operativos
** e-fólio A 2018-19 (mpss.c)
**
** Aluno: 1303334 - Claudia Pires
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>


/*
  Para evitar que os processos escrevam numa posição inexistente do ficheiro,
  inicializa-se o ficheiro com todas as somas iguais a 0
*/
int init_file ( int np )
{
    FILE* file = fopen("data.aux", "w");
    if ( file )
    {
        int array[] = //32 int = 0
        {
            0,0,0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,0,0,
            0,0
        };
        fwrite(array, sizeof(int), np, file );
        fclose(file);
    }
    return 0;
}


/*
  Percorre o ficheiro lendo os resultados parciais de cada processo e retorna a sua acumulação
*/
int calcula ( )
{
    int acumulador = 0 ; //acumulador
    FILE* file = fopen("data.aux", "rb");

    if ( file )
    {
        int parcial = 0;
        while ( fread( &parcial, 1, sizeof(int), file ) == sizeof(int) )
        {
            acumulador += parcial;
        }
        fclose(file);
    }

    return acumulador;
}


/*
  Calcula o resultado esperado que dado pela expressão s(n) = [n(n + 1)(2n + 1)]/6
*/
int aplica_formula ( int n )
{
    return ( n * ( n + 1 ) * ( ( 2 * n ) + 1 ) ) / 6 ;
}


/*
  Processamento parcial, a ser executado pelos np processos.
*/
int mpss_parcial( int i, int np, int n )
{

    int seek   = i * sizeof(int) ;  // Calcula a posição no ficheiro
    int block  =          n / np ;  // O valor do bloco de processamento
    int from   = i++ * block + 1 ;  // O valor inicial do bloco
    int untill = i   * block     ;  // O valor final do bloco

    pid_t pid;

    //caso não seja o último processo
    if ( i < np )
    {
        //tentar efectuar o fork do processo
        if ( ( pid = fork() ) < 0 )
        {
            printf("Erro a efectuar fork do processo 1\n");
            return EXIT_FAILURE;
        }

        //Se conseguiu efectuar o fork
        //verifica se é o processo filho
        if ( pid == 0 )
        {
            //lanca o processamento parcial e termina o processamento
            return mpss_parcial( i, np, n );
        }
    }
    else if ( untill < n ) //caso seja o último processo..
    {
        untill = n; // o último valor do bloco é o ultimo valor  (n)
    }

    // Executa a soma parcial entre o inicio e o fim do bloco
    int soma   = 0;
    for( int j = from; j <= untill; j++)
    {
        soma += ( j * j );
    }

    //Abrir o ficheiro auxiliar em modo de escrita e leitura em modo binario.
    FILE* file = fopen("data.aux", "rb+");
    if ( file )
    {
        //Aponta o ficheiro para a posição do processo
        if ( fseek( file, seek, SEEK_SET ) == 0 )
        {
            // Escreve o resultado parcial no ficheiro
            fwrite(&soma, sizeof(int), 1, file );
        }
        //terminando o filho, fecha o ficheiro auxiliar
        fclose(file);
    }

    // imprime o resultado parcial bem como a informação de processo
    printf( "Processo P%-*d : PID=%*d  PPID=%*d i=[%*d-%*d]  soma=%*d\n", 2, i, 5, getpid(), 5, getppid(), 3, from, 3, untill, 9, soma) ;

    // se não for o último processo, espera que o seu processo filho termine
    if ( i < np )
    {
        int   status;
        pid = wait(&status);
    }
    // Sai com sucesso
    return EXIT_SUCCESS;
}


/*
    Funcção principal do programa, invocada após correcta validação dos parametros de entrada.
*/
int mpss( int np, int n )
{
    printf("Soma do quadrado dos primeiros %d numeros naturais com cadeia de %d processos dedicados\n", n, np);
    printf("Processo inicial P0: PID=%*d  PPID=%*d\n", 5, getpid(), 5, getppid());

    //inicializa o ficheiro que servirá para guardar os resultados parciais da soma
    init_file (np);

    //tenta efectuar o fork do processo
    pid_t pid = fork();
    if ( pid < 0 )
    {
        printf("Erro a efectuar fork do processo 1\n");
        return EXIT_FAILURE;
    }

    //Se conseguiu efectuar o fork
    //verifica se é o processo filho
    if ( pid == 0 )
    {
        //lanca o processamento parcial e termina o processo
        return mpss_parcial( 0, np, n );
    }
    // se for o processo pai, espera pelo termino do processo filho
    int   status;
    pid = wait(&status);


    //calcula o resultado do processamento utilizando o ficheiro auxiliar e a formula
    int calculado = calcula();
    int esperado  = aplica_formula(n);

    //Apresenta os resultados
    printf( "Resultado calculado: %d\n", calculado ) ;
    printf( "Resultado esperado : %d\n",  esperado ) ;

    return EXIT_SUCCESS;
}


/*
  Ponto de entrada do processo mpss, responsavel por validar os parametros de entrada e
  devolver erro caso não sejam válidos.
  Se tudo estiver correcto invoca a função principal mpss(np, n)
*/
int main( int argc, char *argv[] )
{
    if ( argc != 3 ) // mpss np n
    {
        printf("Deve invocar o processo da seguinte forma: ./mpss <np> <n>\n");
        printf(" np - nº de processos dedicados ao cálculo do somatório, com 1 ≤ np ≤ 32\n");
        printf("  n - nº de termos do somatório, com np ≤ n ≤999\n");

        return EXIT_FAILURE;
    }

    // sscanf para converter o parametro de entrada de string para inteiro
    // nao foi utilizado o 'atoi' pois nao seria possivel fazer controle de erros
    // argv[0] - Nome do executavel
    // argv[1] - Argumento np
    // argv[2] - Argumento n

    // Validacao do np
    int np;
    if ( sscanf( argv[1], "%d", &np ) != 1 )      
    {
        printf("Erro a converter np para inteiro\n");
        return EXIT_FAILURE;
    }
    if ( np < 1 || np > 32 )
    {
        printf("Valor incorrecto de np [1 <= np <= 32]\n");
        return EXIT_FAILURE;
    }

    // Validacao do n
    int n;
    if ( sscanf( argv[2], "%d", &n ) != 1 )
    {
        printf("Erro a converter n para inteiro\n");
        return EXIT_FAILURE;
    }
    if (  n < np || n > 999 )
    {
        printf("Valor incorrecto de n [np <= n <= 999]\n");
        return EXIT_FAILURE;
    }

    return mpss(np, n);
}

// fim ficheiro "mpss.c"
